from pandas.core.common import is_bool_indexer

from xl_link.xlsxwriter.utility import xl_rowcol_to_cell, xl_cell_to_rowcol


def is_int_type(i):
    try:
        return int(i) == i
    except TypeError:
        return False
    except ValueError:
        return False


def fill_slice(holey_slice):
    return slice(holey_slice.start if holey_slice.start is not None else 0,
                 holey_slice.stop if holey_slice.stop is not None else -1,
                 holey_slice.step if holey_slice.step is not None else 1)

def to_series(cell_or_range):
    return '=' + cell_or_range


class XLCell:
    """
    Represents the location of a cell in an Excel spreadsheet

    Parameters
    ----------
    row : int
        corresponding to cell row number (0 indexed)
    col : int
        corresponding to cell col number (0 indexed)
    sheet: str
        corresponding to the sheet the cell is in, default='Sheet1'

    Examples
    --------
    >>> XLCell(0, 5, 'Accounts')
        <XLCell: F1>

    >>> XLCell.from_cell('F1', 'Accounts')
        <XLCell: F1>

    >>> XLCell.from_fcell("'Accounts'!F1")
        <XLCell: F1>

    Notes
    -----
    sheet can be set to None if desired, keep in mind it defaults to 'Sheet1', mostly for convenience.

    row, col and sheet are attributes that can be accessed and set directly.

    All methods are designed to return new objects, preserving the state of self.

    See Also
    --------
    XLCell.from_cell
    XLCell.from_fcell
    """

    def __init__(self, row, col, sheet='Sheet1'):
        self.sheet = sheet
        self.row = row
        self.col = col

    @classmethod
    def from_cell(cls, cell, sheet="Sheet1"):
        """
        Alternative constructor

        Paramters
        ---------
        cell : str
            cell in excel notation (e.g. 'A1'), representing cell location.
        sheet : sheet
            sheet cell belongs to, default='Sheet1'

        Returns
        -------
        XLCell
            Initialised XLCell
        """
        return cls(*xl_cell_to_rowcol(cell), sheet)

    @classmethod
    def from_fcell(cls, fcell):
        """
        Alternative constructor

        Paramters
        ---------
        fcell: str
            cell in excel formula notation (e.g. "'Sheet1'!A1"), representing cell location.

        Returns
        -------
        XLCell
            Initialised XLCell
        """
        try:
            sheet, cell = fcell.split('!')
        except ValueError:
            raise ValueError("""Could not parse fcell: {}, is your fcell in the form "'Sheet1'!A1:B2"?""".format(fcell))
        return cls(*xl_cell_to_rowcol(cell), sheet)

    @property
    def cell(self):
        """
        Gets the cell location str.

        Returns
        -------
        str
            cell location in excel notation

        Examples
        --------
        >>> cell = XLCell(0, 5, 'Accounts')
        >>> cell.cell
            'F1'
        """
        return xl_rowcol_to_cell(self.row, self.col)

    @property
    def fcell(self):
        """
        Gets the cell location for use in excel formulas.

        Returns
        -------
        str
            cell location in excel notation, in '{sheet}'!{cell} form

        Examples
        --------
        >>> cell = XLCell(0, 5, 'Accounts')
        >>> cell.fcell
            "'Accounts'!F1"
        """
        return "'{}'!{}".format(self.sheet, self.cell)

    @property
    def rowcol(self):
        """
        Coordinates of cell

        Returns
        -------
        row, col : tuple(int, int)
            (row, col) corresponding to location of cell
        """
        return self.row, self.col

    def range_between(self, other):
        """
        Get XLRange between this cell and other.

        Parameters
        ----------
        other : XLCell
            other XLCell that makes up stop of excel range

        Returns
        -------
        XLRange
            range between this cell and other

        Example
        -------
        >>> start = XLCell.from_cell('A1')
        >>> stop = XLCell.from_cell('B7')
        >>> start.range_between(stop)
            <XLRange: A1:B7>

        """
        return XLRange(self, other)

    def __sub__(self, other):
        """
        Alias for XLCell.range_between
        """
        return XLRange(self, other)

    def __eq__(self, other):
        """
        Equality check.

        Parameters
        ----------
        other : object
            to compare for equality.

        Notes
        -----
        other can be another XLCell or str, if a str is provided, XLCell.from_fcell will be called on other, and then
        the comparison made. Comparisons between XLCells first compares sheets, and then for equal rows and columns.
        """
        if isinstance(other, str):
            return self == XLCell.from_fcell(other)

        if isinstance(other, XLCell):
            if self.sheet != other.sheet:
                return False
            return self.row == other.row and self.col == other.col

        return False

    def __repr__(self):
        return "<XLCell: {}>".format(self.fcell)

    def copy(self):
        """
        Returns
        -------
        XLCell
            copy of self
        """
        return XLCell(self.row, self.col, self.sheet)

    def translate(self, row, col):
        """
        Returns new XLCell with translation applied

        Parameters
        ----------
        row : int
            corresponding to movement in row direction (+ve down the spreadsheet)
        col : int
            corresponding to movement in col direction (+ve right across the spreadsheet)

        Returns
        -------
        XLCell
         new cell with translation applied
        """
        cell = self.copy()
        cell.row += row or 0
        cell.col += col or 0
        return cell

    def __hash__(self):
        return hash((self.row, self.col, self.sheet))

class XLRange:
    """
    Represents a range in an Excel spreadsheet

    Parameters
    ----------
    start : XLCell
        corresponding to start cell
    stop : XLCell
        corresponding to stop cell

    Examples
    --------
    Using __init__

    >>> start = XLCell.from_cell('A1')
    >>> stop = XLCell.from_cell('B7')
    >>> XLRange(start, stop)
        <XLRange: A1:B7>

    Notes
    -----
    Upon initialisation, if start and stop are not from the same sheet, an assertion error is raised.

    start and stop are attributes that can be accessed and modified directly.

    All methods are designed to return new objects, preserving the state of self.

    See Also
    --------
    XLRange.from_range
    XLRange.from_frange
    """
    def __init__(self, start, stop):
        assert start.sheet == stop.sheet, "start and stop must be in the same sheet"
        self.sheet = start.sheet
        self.start = start.copy()
        self.stop = stop.copy()

    @classmethod
    def from_range(cls, range, sheet='Sheet1'):
        """
        Alternative constructor

        Parameters
        ----------

        range : str
            str in excel notation (e.g. 'A1:B20'), representing excel range.
        sheet : str
            sheet range belongs to, default='Sheet1'

        Returns
        -------
        XLRange
            initialised XLRange
        """
        start, stop = range.split(':')
        return cls(XLCell.from_cell(start, sheet), XLCell.from_cell(stop, sheet))

    @classmethod
    def from_frange(cls, frange):
        """
        Alternative constructor

        Parameters
        ----------
        frange : str
            str in excel formula notation (e.g. "'Sheet1'!A1:B20"), representing excel range.

        Returns
        -------
        XLRange
            initialised XLRange
        """
        sheet, range = frange.split('!')
        return cls.from_range(range, sheet)

    @property
    def range(self):
        """
        Gets the Excel range this object represents

        Returns
        -------
        str
            representing this range in excel notation

        Examples
        --------
        >>> start = XLCell.from_cell('A1')
        >>> stop = XLCell.from_cell('B7')
        >>> range = XLRange(start, stop)
        >>> range.range
            'A1:B7'
        """
        return "{}:{}".format(self.start.cell, self.stop.cell)

    @property
    def frange(self):
        """
        Gets the Excel range this object represents, for use in excel formulas

        Returns
        -------
        str
            representing this range in excel notation for use in excel formulas (e.g. "'{sheet}'!{start}:{stop}"

        Examples
        --------
        >>> start = XLCell.from_cell('A1')
        >>> stop = XLCell.from_cell('B7')
        >>> range = XLRange(start, stop)
        >>> range.frange
            "'Sheet1'!A1:B7"
        """
        return "'{}'!{}".format(self.sheet, self.range)

    @property
    def rowcol_rowcol(self):
        """
        Coordinates of start and stop

        Returns
        -------
        start.row, stop.col : tuple
            start position in rows and column num (start.row, start.col)
        stop.row, stop.col : tuple
            stop position in rows and column num (stop.row, stop.col)
        """
        return self.start.rowcol, self.stop.rowcol

    @property
    def shape(self):
        """
        Returns
        -------
        tuple
            representing shape of XLRange in form (height, width)
        """
        return self.stop.row - self.start.row + 1, self.stop.col - self.start.col + 1

    @property
    def is_row(self):
        return self.shape[0] == 1

    @property
    def is_col(self):
        return self.shape[1] == 1

    @property
    def is_1D(self):
        return self.is_row or self.is_col

    @property
    def is_2D(self):
        return not self.is_1D

    def __repr__(self):
        return "<XLRange: {}>".format(self.frange)

    def __len__(self):
        """
        Length of self

        Notes
        -----
        Will raise TypeError unless self is 1 dimensional
        """
        if self.is_col:
            return self.shape[0]
        if self.is_row:
            return self.shape[1]
        raise TypeError("length is only defined for 1D ranges")

    def __iter__(self):
        """
        iterate over each XLCell with self, if self is a row or column

        Notes
        -----
        Will raise TypeError if self is not 1 dimensional
        """
        if self.is_1D:
            for x in range(len(self)):
                yield self[x]
        else:
            raise TypeError("Can only iterate over 1D ranges")

    def __getitem__(self, key):
        """
        Get subsection of self using given key

        Parameters
        ----------
        key : int or slice or boolian indexer or tuple of 2 ints/ slices!

        Returns
        -------
        XLCell or XLRange
            result after selection using key applied. Designed to emulate behaviour of Pandas Indexes.

        Notes
        -----
        int, single slice and boolean indexers can only be used on one dimensional XLRanges.
        """
        if is_int_type(key):

            if not self.is_1D:
                raise TypeError("Can only do integer lookups on 1D ranges")

            elif key < 0:
                key += len(self)

            if self.is_row:
                return self.start.translate(0, key)
            else:
                return self.start.translate(key, 0)

        elif isinstance(key, slice):
            key = fill_slice(key)
            if key.step != 1:
                raise TypeError("Can only slice with step equal to 1")

            start, stop = key.start, (key.stop)

            return self[start] - self[stop]

        elif is_bool_indexer(key):

            if not self.is_1D:
                raise TypeError("Can only use Boolean indexers on 1D ranges")

            true_positions = list(i for i, bool_ in enumerate(key) if bool_ == True)

            start = min(true_positions)
            stop = max(true_positions)

            step_is_1 = (stop - start) / len(true_positions) == 1

            if not step_is_1:
                raise TypeError("Bool indexers can't have any holes in (i.e. equivalent as slice must have step=1)")

            return self[start:stop]

        elif len(key) == 2:

            if self.is_1D:
                raise TypeError("2 Parameter indexing only available on 2D ranges")

            row_slice, col_slice = key

            if is_int_type(row_slice) and is_int_type(col_slice):

                if row_slice < 0:
                    row_slice += self.shape[0]
                if col_slice < 0:
                    col_slice += self.shape[1]

                return self.start.translate(row_slice, col_slice)

            elif isinstance(row_slice, slice) and isinstance(col_slice, slice):
                row_slice = fill_slice(row_slice)
                col_slice = fill_slice(col_slice)

                row_stop = row_slice.stop
                if row_slice.stop < 0:
                    row_stop += self.shape[0]

                col_stop = col_slice.stop
                if col_slice.stop < 0:
                    col_stop += self.shape[1]

                return XLRange(self.start.translate(row_slice.start, col_slice.start),
                               self.start.translate(row_stop, col_stop))

        raise TypeError("Expecting tuple of slices, boolean indexer, or an index or a slice if 1D, not {}".format(key))

    def iterrows(self):
        """
        Iterate over each row of self, yields each row as XLRange

        Yields
        -------
        XLRange
            corresponding to current row
        """
        if not self.is_2D:
            raise TypeError("Can only call iterrows on 2D ranges")

        for x in range(self.shape[0]):
            yield self[x, 0] - self[x, -1]

    def __eq__(self, other):
        return self.start == other.start and self.stop == other.stop

    def copy(self):
        return self.start.copy() - self.stop.copy()

    def __hash__(self):
        return hash((self.start.row, self.start.col, self.stop.row, self.stop.row, self.sheet))

    def translate(self, row, col):
        """
        Translates whole range by row, col.

        Parameters
        ----------
        row : int
            corresponding to movement in row direction (+ve down the spreadsheet)
        col : int
            corresponding to movement in col direction (+ve right across the spreadsheet)

        Returns
        -------
        XLRange
            new with translation applied

        Notes
        -----
        This moves the whole range, and cannot be used to change the shape of self.
        """
        new = self.copy()

        new.start = new.start.translate(row, col)
        new.stop = new.stop.translate(row, col)

        return new




from abc import abstractmethod
import importlib
from xl_link.xl_types import to_series

from distutils.version import StrictVersion

imported = {}

s = to_series

SINGLE_CATEGORY_CHARTS = ['bar', 'area', 'doughnut', 'line', 'pie', 'radar', 'stock', 'column']
CATEGORIES_REQUIRED_CHARTS = ['scatter']

MIN_VERSIONS = {'xlsxwriter': '0.9',
                'openpyxl': '2.4'}


def check_engine_compatible(engine):
    min_version = MIN_VERSIONS.get(engine.__name__, 0)
    if StrictVersion(engine.__version__) < StrictVersion(min_version):
        raise ImportError(("excel writer engine not met version requirements for xl_link:"
                           "{} < {}").format(engine.__version__ < min_version))


def ensure_list(specifier):
    """
    if specifier isn't a list or tuple, makes specifier into a list containing just specifier
    """
    if not isinstance(specifier, list) and not isinstance(specifier, tuple):
        return [specifier,]

    return specifier


class AbstractChartWrapper:
    """
    Wraps around excel writer module, for use by create_chart.
    """

    def __init__(self, book, type_, subtype=None, engine=None):
        """
        Abstract: should initialise Chart Wrapper class. Should initialise chart object, and set as self.chart.

        Parameters
        ----------
        book : object
            pd.ExcelWriter.book object typical represents whole excel file, and needed to creat new chart
        type_ : str
            type of chart desired
        subtype : str
            subtype of chart desired
        engine : package
            corresponding to engine used (usefull for fetching objects from that package)
        """
        self.type_ = type_
        self.book = book
        self.subtype = subtype
        self.engine = engine
        self.chart = None

    @abstractmethod
    def add_series(self, name, values, categories=None):
        """
        Abstract: should add a series to chart.

        Parameters
        ----------
        name : str
            representing the name of the series
        values : XLRange
            represending values of series
        categories : XLRange
            representing values of categories
        """
        pass

    @property
    @abstractmethod
    def x_axis_name(self):
        pass

    @x_axis_name.setter
    @abstractmethod
    def x_axis_name(self, value):
        pass

    @property
    @abstractmethod
    def y_axis_name(self):
        pass

    @y_axis_name.setter
    @abstractmethod
    def y_axis_name(self, value):
        pass

    @property
    @abstractmethod
    def title(self):
        pass

    @title.setter
    @abstractmethod
    def title(self, value):
        pass


class XlsxWriterChartWrapper(AbstractChartWrapper):

    def __init__(self, book, type_, subtype):
        super().__init__(book, type_, subtype)

        self.chart = book.add_chart({'type': type_, 'subtype': subtype} if subtype else {'type': type_})

    def add_series(self, name, values, categories=None):
        values = s(values.frange)

        kwargs = {'name': name, 'values': values}

        if categories:
            kwargs['categories'] = s(categories.frange)

        self.chart.add_series(kwargs)

    @property
    def x_axis_name(self):
        return self.chart.x_axis['name']

    @x_axis_name.setter
    def x_axis_name(self, value):
        self.chart.x_axis['name'] = value

    @property
    def y_axis_name(self):
        return self.chart.y_axis['name']

    @y_axis_name.setter
    def y_axis_name(self, value):
        self.chart.y_axis['name'] = value

    @property
    def title(self):
        return self.chart.title_name

    @title.setter
    def title(self, value):
        self.chart.title_name = value


def type_to_openpyxl_chart_name(type_):
    return type_.title() + 'Chart'


class OpenPyXLChartWrapper(AbstractChartWrapper):

    def __init__(self, book, type_, subtype, openpyxl):
        super().__init__(book, type_, subtype, openpyxl)

        self.chart = getattr(openpyxl.chart, type_to_openpyxl_chart_name(type_))()
        if subtype:
            self.chart.type = subtype

    def add_series(self, name, values, categories=None):
        Series = self.engine.chart.series_factory.SeriesFactory

        if categories is None:
            series = Series(values.frange, title=name)
            self.chart.append(series)
        elif self.type_ in SINGLE_CATEGORY_CHARTS:
            values.start.row += -1 # To include top cell as name
            self.chart.add_data(values.frange, titles_from_data=True)
            self.chart.set_categories(categories.frange)
        else:
            series = Series(values.frange, xvalues=categories.frange, title=name)
            self.chart.append(series)

    @property
    def x_axis_name(self):
        return self.chart.x_axis.title

    @x_axis_name.setter
    def x_axis_name(self, value):
        self.chart.x_axis.title = value

    @property
    def y_axis_name(self):
        return self.chart.y_axis.title

    @y_axis_name.setter
    def y_axis_name(self, value):
        self.chart.y_axis.title = value

    @property
    def title(self):
        return self.chart.title

    @title.setter
    def title(self, value):
        self.chart.title = value

def create_chart(workbook, engine, type_, values, categories, names, subtype=None,
                 title=None, x_axis_name=None, y_axis_name=None):
    """
    Create a chart object corresponding to given engine, within sheet.

    Parameters
    ----------
    workbook : object
        to insert chart into, either XlsxWriter.Workbooks or openpyxl Workbooks.
    engine : str
        representing engine to use, either XlsxWriter.Workbooks or openpyxl Workbooks.
    type_ : str
        representing chart type, see engine docs for options e.g. 'line'.
    values : XLRange or sequence of XLRanges
        use as values for each series/ data set (Excel equivalent to y data).
    categories : XLRange or sequence of XLRanges
        use as categories for each series/ data set (Excel equivalent to x data).
    names : str/ sequence of str
        use as name for each series/ data set (Excel uses this to label each dataset).
    subtype : str
        representing chart subtype, see engine docs for details.

    Returns
    -------
    chart : object
        populated chart object corresponding to engine's chart type.
    """
    if 'openpyxl' in engine:
        engine = 'openpyxl' # Cuz pandas appends version to engine name

    values = ensure_list(values)
    categories = ensure_list(categories)
    names = ensure_list(names)

    if engine not in imported:
        engine_mod = importlib.import_module(engine)
        check_engine_compatible(engine_mod)
        imported[engine] = engine_mod

    if engine == 'xlsxwriter':
        chart = XlsxWriterChartWrapper(workbook, type_, subtype)
    elif engine == 'openpyxl':
        chart = OpenPyXLChartWrapper(workbook, type_, subtype, imported[engine])
    else:
        raise TypeError("Couldn't find chart wrapper for {}".format(engine))

    if len(categories) == 1 and len(categories) < len(values):
        categories *= len(values)

    for name, category, value in zip(names, categories, values):
        chart.add_series(name, value, category)

    if title:
        chart.title = title
    if x_axis_name:
        chart.x_axis_name = x_axis_name
    if y_axis_name:
        chart.y_axis_name = y_axis_name

    return chart.chart


import pandas as pd

try:
    from pandas.io.formats.excel import ExcelFormatter
except ImportError:
    from pandas.formats.format import ExcelFormatter

from pandas.io.common import _stringify_path

from .xl_types import XLCell
from .chart_wrapper import create_chart, SINGLE_CATEGORY_CHARTS, CATEGORIES_REQUIRED_CHARTS


def get_xl_ranges(frame_index, frame_columns,
                  sheet_name='Sheet1',
                  columns=None,
                  header=True,
                  index=True,
                  index_label=None,
                  startrow=0,
                  startcol=0,
                  merge_cells=True):
    """
    Deduces location of data_range, index_range and col_range within excel spreadsheet, given the parameters provided.

    Does not require an actual DataFrame, which could be useful!

    Parameters
    ----------
    frame_index: Pandas Index or Array-like
        to determine location of index within spreadsheet.
    frame_columns: Pandas Index or Array-like
        used to determine location of column within spreadsheet.
    excel_writer : string or ExcelWriter

    sheet_name : str
        default ‘Sheet1’, Name of sheet which will contain DataFrame
    columns : sequence
        optional, Columns to write
    header : bool or list of strings,
        default True Write out the column names. If a list of strings is given it is assumed to be aliases for the column names
    index : bool
        default True. Write row names (index)
    index_label : str or sequence
        default None. Column label for index column(s) if desired. If None is given, and header and index are True, then the index names are used. A sequence should be given if the
        DataFrame uses MultiIndex.
    startrow : int
        upper left cell row to dump data frame
    startcol : int
        upper left cell column to dump data frame
    merge_cells : bool
        default True. Write MultiIndex and Hierarchical Rows as merged cells.

    Returns
    -------
    data_range, index_range, col_range : XLRange
        Each range represents where the data, index and columns can be found on the spreadsheet
    empty_f : DatFrame
        an empty DataFrame with matching Indices.
    """

    empty_f = pd.DataFrame(index=frame_index, columns=frame_columns)

    formatter = ExcelFormatter(empty_f,
                               cols=columns,
                               header=header,
                               index=index,
                               index_label=index_label,
                               merge_cells=merge_cells)

    excel_header = list(formatter._format_header())
    col_start, col_stop = excel_header[0], excel_header[-1]

    col_start_cell = XLCell(col_stop.row + startrow, col_start.col + startcol, sheet_name)
    col_stop_cell = XLCell(col_stop.row + startrow, col_stop.col + startcol, sheet_name)

    if isinstance(empty_f.index, pd.MultiIndex):
        col_start_cell = col_start_cell.translate(0, 1)

    col_range = col_start_cell - col_stop_cell

    body = list(formatter._format_body())

    if empty_f.index.name or index_label:
        body.pop(0)  # gets rid of index label cell that comes first!

    index_start_cell = XLCell(body[0].row + startrow, body[0].col + startcol + empty_f.index.nlevels - 1, sheet_name)
    index_stop_cell = XLCell(body[-1].row + startrow, body[0].col + startcol + empty_f.index.nlevels - 1, sheet_name)

    index_range = index_start_cell - index_stop_cell

    data_start_cell = XLCell(index_start_cell.row, col_start_cell.col, sheet_name)
    data_stop_cell = XLCell(index_stop_cell.row, col_stop_cell.col, sheet_name)

    data_range = data_start_cell - data_stop_cell

    return data_range, index_range, col_range, empty_f


def write_frame(f, excel_writer, to_excel_args=None):
    """
    Write a Pandas DataFrame to excel by calling to_excel, returning an XLMap, that can be used to determine
    the position of parts of f, using pandas indexing.

    Parameters
    ----------
    f : DataFrame
        Frame to write to excel
    excel_writer : str or ExcelWriter
        Path or existing Excel Writer to use to write frame
    to_excel_args : dict
        Additional arguments to pass to DataFrame.to_excel, see docs for DataFrame.to_excel

    Returns
    -------
    XLMap :
        Mapping that corresponds to the position in the spreadsheet that frame was written to.
    """
    xlf = XLDataFrame(f)

    return xlf.to_excel(excel_writer, **to_excel_args)


def _mapper_to_xl(value):
    """
    Convert mapper frame result to XLRange or XLCell
    """
    if isinstance(value, XLCell):
        return value

    if isinstance(value, pd.Series):
        return value.values[0] - value.values[-1]

    if isinstance(value, pd.DataFrame):
        return value.values[0, 0] - value.values[-1, -1]

    raise TypeError("Could not conver {} to XLRange or XLCell".format(value))


class _SelectorProxy:
    """
    Proxy object that intercepts calls to Pandas DataFrame indexers, and re-interprets result into excel locations.

    Parameters
    ----------
    mapper_frame: DataFrame
        with index the same as the DataFrame it is representing, however, each cell contains
        the location they sit within the spreadsheet.
    selector_name: str
        name of the indexer SelectorProxy is emulating, i.e. loc, iloc, ix, iat or at

    Notes
    -----
    Only implements __getitem__ behaviour of indexers.
    """

    def __init__(self, mapper_frame, selector_name):
        self.mapper_frame = mapper_frame
        self.selector_name = selector_name

    def __getitem__(self, key):
        val = getattr(self.mapper_frame, self.selector_name)[key]

        return _mapper_to_xl(val)


class XLMap:
    """
    An object that maps a Pandas DataFrame to it's positions on an excel spreadsheet.

    Provides access to basic pandas indexers - __getitem__, loc, iloc, ix, iat and at.
    These indexers are modified such that they return the cell/ range of the result.

    The idea is should make using the data in spreadsheet easy to access, by using Pandas indexing syntax.
    For example can be used to create charts more easily (see example below).

    Notes
    -----

    Recommended to not be created directly, instead via, XLDataFrame.to_excel.

    XLMap can only go 'one level deep' in terms of indexing, because each indexer always returns either an XLCell,
    or an XLRange. The only workaround is to reduce the size of your DataFrame BEFORE you call write_frame.
    This limitation drastically simplifies the implementation. Examples of what WON'T WORK:

    >>> xlmap.loc['Mon':'Tues', :].index
        AttributeError: 'XLRange' object has no attribute 'index'

    >>> xlmap.index['Mon':'Tues'] # Doesn't work because index is not a Pandas Index, but an XLRange.
        TypeError: unsupported operand type(s) for -: 'str' and 'int'

    Parameters
    ----------
    data_range, index_range, column_range : XLRange
     that represents the region the DataFrame's data sits in.
    f : DataFrame
     that has been written to excel.

    Examples
    --------
    >>> calories_per_meal = XLDataFrame(columns=("Meal", "Mon", "Tues", "Weds", "Thur"),
                                         data={'Meal': ('Breakfast', 'Lunch', 'Dinner', 'Midnight Snack'),
                                               'Mon': (15, 20, 12, 3),
                                               'Tues': (5, 16, 3, 0),
                                               'Weds': (3, 22, 2, 8),
                                               'Thur': (6, 7, 1, 9)})
    >>> calories_per_meal.set_index("Meal", drop=True, inplace=True)

    Write to excel

    >>> writer = pd.ExcelWriter("Example.xlsx", engine='xlsxwriter')
    >>> xlmap = calories_per_meal.to_excel(writer, sheet_name="XLLinked") # returns the XLMap

    Create chart with XLLink

    >>> workbook = writer.book
    >>> xl_linked_sheet = writer.sheets["XLLinked"]
    >>> xl_linked_chart = workbook.add_chart({'type': 'column'})
    >>> for time in calories_per_meal.index:
    >>>     xl_linked_chart.add_series({'name': time,
                                        'categories': proxy.columns.frange,
                                        'values': proxy.loc[time].frange})
    """

    def __init__(self, data_range, index_range, column_range, f, writer=None):
        self.index = index_range
        self.columns = column_range

        self.data = data_range
        self._f = f.copy()

        self.writer = writer
        self.book = writer.book
        self.sheet = writer.sheets[self.index.sheet]

        self._mapper_frame = f.copy().astype(XLCell)

        x_range = self._f.index.size
        y_range = self._f.columns.size

        for x in range(x_range):
            for y in range(y_range):
                self._mapper_frame.values[x, y] = data_range[x, y]

    @property
    def f(self):
        """
        for convenience provides read-only access to the DataFrame originally written to excel.
        """
        return self._f

    def __repr__(self):
        return "<XLMap: index: {}, columns: {}, data: {}>".format(self.index, self.columns, self.data)

    def create_chart(self, type_='scatter',
                     values=None, categories=None, names=None,
                     subtype=None,
                     title=None, x_axis_name=None, y_axis_name=None):
        """
        Create excel chart object based off of data within the Frame.

        Parameters
        ----------
        type_ : str
            Type of chart to create.
        values, categories, names: str or list or tuple
            label or list of labels to corresponding to column to use as values and categories and names for each series in chart.
        subtype : str
            subtype of type, only available for some chart types e.g. bar, see Excel writing package for details
        title : str
            chart title
        x_axis_name : str
            used as label on x_axis
        y_axis_name : str
            used as label on y_axis

        Returns
        -------

        Chart object corresponding to the engine selected

        """

        if names is None and categories is None:
            names = tuple(name for name in self.f.columns.values)
        elif names is None and isinstance(categories, (str, int, list, tuple)):
            names = categories
        elif isinstance(names, (list, tuple)):
            names = tuple(self.f[names].columns.values)
        else:
            raise TypeError("Couldn't understand names input: " + names)

        if values is None:
            values = tuple(self[value] for value in self.f.columns)
        elif isinstance(values, list) or isinstance(values, tuple):
            values = tuple(self[value] for value in values)
        else:
            values = self[values]

        if categories is None and (type_ in SINGLE_CATEGORY_CHARTS and isinstance(values, tuple)) or \
                        type_ in CATEGORIES_REQUIRED_CHARTS:
            categories = self.index # Default, use x as index
        elif categories is None:
            pass
        elif isinstance(categories, (list, tuple)):
            categories = list(self[category] for category in categories)
        else:
            categories = self[categories]

        return create_chart(self.book, self.writer.engine, type_,
                            values, categories, names,
                            subtype, title,
                            x_axis_name, y_axis_name)

    def __getitem__(self, key):
        """
        Emulates DataFrame.__getitem__ (DataFrame[key] syntax), see Pandas DataFrame indexing for help on behaviour.

        Will return the location of the columns found, rather than the underlying data.

        Parameters
        ----------
        key : hashable or array-like
            hashables, corresponding to the names of the columns desired.

        Returns
        -------
        XLRange :
            corresponding to position of found colummn(s) within spreadsheet

        Example
        -------
        >>> xlmap['Col 1']
            <XLRange: B2:B10>
        """
        val = self._mapper_frame[key]

        return _mapper_to_xl(val)

    @property
    def loc(self):
        """
        Proxy for DataFrame.loc, see Pandas DataFrame loc help for behaviour.

        Will return location result rather than underlying data.

        Returns
        -------
        XLCell or XLRange
            corresponding to position of DataFrame, Series or Scalar found within spreadsheet.

        Example
        -------
        >>> xlmap.loc['Tues']
            <XLRange: A2:D2>
        """
        return _SelectorProxy(self._mapper_frame, 'loc')

    @property
    def iloc(self):
        """
        Proxy for DataFrame.iloc, see Pandas DataFrame iloc help for behaviour.

        Will return location result rather than underlying data.

        Returns
        -------
        XLCell or XLRange
            corresponding to position of DataFrame, Series or Scalar found within spreadsheet.

        Example
        -------
        >>> xlmap.iloc[3, :]
            <XLRange: A2:D2>
        """
        return _SelectorProxy(self._mapper_frame, 'iloc')

    @property
    def ix(self):
        """
        Proxy for DataFrame.ix, see Pandas DataFrame ix help for behaviour. (That said this is deprecated since 0.20!)

        Will return location result rather than underlying data.

        Returns
        -------
        XLCell or XLRange
            corresponding to position of DataFrame, Series or Scalar found within spreadsheet.


        Example
        -------
        >>> xlmap.ix[3, :]
            <XLRange A2:D2>
        """
        return _SelectorProxy(self._mapper_frame, 'ix')

    @property
    def iat(self):
        """
        Proxy for DataFrame.iat, see Pandas DataFrame iat help for behaviour.

        Will return location result rather than underlying data.

        Returns
        -------
        XLCell
            location corresponding to position value within spreadsheet.

        Example
        -------
        >>> xlmap.iat[3, 2]
            <XLCell C3>
        """
        return _SelectorProxy(self._mapper_frame, 'iat')

    @property
    def at(self):
        """
        Proxy for DataFrame.at, see Pandas DataFrame at help for behaviour.

        Will return location result rather than underlying data.

        Returns
        -------
        XLCell
            location corresponding to position value within spreadsheet.


        Example
        -------
        >>> xlmap.at["Mon", "Lunch"]
            <XLCell: C3>
        """
        return _SelectorProxy(self._mapper_frame, 'at')


class XLDataFrame(pd.DataFrame):
    """
    Monkeypatched DataFrame modified by xl_link!

    Changes:
    --------

    * to_excel modified to return an XLMap.

    * XLDataFrame._constructor set to XLDataFrame -> stops reverting to normal DataFrame

    Notes
    -----

    Conversions from this DataFrame to Series or Panels will return regular Panels and Series,
    which will convert back into regular DataFrame's upon expanding/ reducing dimensions.

    See Also
    --------
    Pandas.DataFrame
    """

    @property
    def _constructor(self):
        return XLDataFrame

    def to_excel(self, excel_writer, sheet_name='Sheet1', na_rep='',
                 float_format=None, columns=None, header=True, index=True,
                 index_label=None, startrow=0, startcol=0, engine=None,
                 merge_cells=True, encoding=None, inf_rep='inf', verbose=True,
                 **kwargs):
        """

        Monkeypatched DataFrame.to_excel by xl_link!

        Changes:
        --------

        Returns
        -------

        XLMap
            corresponding to position of frame as it appears in excel (see XLMap for details)

        See Also
        --------

        Pandas.DataFrame.to_excel for info on parameters

        """

        if isinstance(excel_writer, pd.ExcelWriter):
            need_save = False
        else:
            excel_writer = pd.ExcelWriter(_stringify_path(excel_writer), engine=engine)
            need_save = True

        super().to_excel(excel_writer, sheet_name=sheet_name, na_rep=na_rep,
                 float_format=float_format, columns=columns, header=header, index=index,
                 index_label=index_label, startrow=startrow, startcol=startcol, engine=engine,
                 merge_cells=merge_cells, encoding=encoding, inf_rep=inf_rep, verbose=verbose,
                 **kwargs)

        if need_save:
            excel_writer.save()

        data_range, index_range, col_range, _ = get_xl_ranges(self.index, self.columns,
                                                              sheet_name=sheet_name,
                                                              columns=columns,
                                                              header=header,
                                                              index=index,
                                                              index_label=index_label,
                                                              startrow=startrow,
                                                              startcol=startcol,
                                                              merge_cells=merge_cells)
        f = self.copy()

        if isinstance(columns, list) or isinstance(columns, tuple):
            f = f[columns]

        return XLMap(data_range, index_range, col_range, f, writer=excel_writer)


from distutils.core import setup
from setuptools.config import read_configuration

setup(name='xl_link',
    version='0.12',
    description='Pandas Magic for Excel spreadsheets',
    author='Hugh Ramsden',
    url='https://github.com/0Hughman0/xl_link',
    download_url="https://github.com/0Hughman0/xl_link/archive/0.12.tar.gz",
    packages=['xl_link', 'xl_link.xlsxwriter', 'xl_link.xl_types'],
    license='MIT',
    install_requires='pandas>=0.19',
    extras_require={'xlsxwriter': 'xslxwriter>=0.9',
                    'openpyxl': 'openpyxl>=2.4'},
    python_require=">=3.4")

conf_dict = read_configuration('./setup.cfg')